
##########################################################################
# Jill Rubinstein
# MBB 752 Spring 2005 
# Final Project 
#
# A simple HMM for gene prediction.
# (Written in R)
#
# Currently consists of four states: coding, non-coding, start codon, 
# stop codon.
#
# To run: call emiter(# of iterations, sequence)
#
# Implements EM algorithm to iteratively improve the HMM's estimated 
# transition, emission, and initial probabilities. 
#
# Implements Viterbi algorithm to reconstruct best path through model,
# i.e. best gene structure for a given input sequence.
# 
# Output: a string of the most probable state for each codon from start 
# to end of input sequence.
# 
# Improvements: train states from large dataset
# Add states for splice site donor, splice site acceptor, forward and 
# reverse strands. Add convenience functions (ex. input format)
#
##########################################################################

#3)Forward-backward algorithm

#Forward algorithm
forward<-function(O,A,B,Pi,S,N){
	alphas <- matrix(c(1:(S*N)),S,N,byrow=TRUE);
	#initialize
	for(i in 1:S){
		alphas[i,1] <- Pi[i] * B[i,O[1]];
	}
	#induction
	for(j in 2:N){
		for(k in 1:S){
			alphas[k,j]<-0;
			for(l in 1:S){		#each possible previous state
				alphas[k,j]<-alphas[k,j]+(alphas[l,j-1]*A[l,k]);
			}
			alphas[k,j]<-(alphas[k,j]* B[k,O[j]]);
		}		
	}
	alphas;
}

#Backward algorithm
backward<-function(O,A,B,Pi,S,N){
	betas <- matrix(c(1:(S*N)),S,N,byrow=TRUE);
	#initialize
	for(i in 1:S){
		betas[i,1] <- Pi[i]* B[i,O[1]];
		betas[i,N] <- 1;
	}
	#induction
	for(j in (N-1):1){
		for(k in 1:S){
			betas[k,j]<-0;
			for(l in 1:S){
				betas[k,j]<-betas[k,j]+(betas[l,j+1]*A[k,l]*B[l,O[j+1]]);
			}
		}
		
	}
	betas;
}

#Calculate log likelihood
forwardbackward<-function(al,be){
	li<-0;
	for(i in 1:S){
		li<-li+(al[i,N]*be[i,N]);
	}
	li<-log10(li);
	li;
}

#method for EM iteration including Viterbi state sequence reproduction
emiter<-function(iter,se){   
   
	S <- 4; 				#number of states
						#(1=exon, 2=intron, 3=start codon, 4=stop codon)
	
	#se <- obs;				#sequence
	
	#read in codons
	z <- 1;
	x <- 1;
	while(substr(se,z+2,z+2) != ""){
		if(x==1){
			O <- c(substr(se,z,z+2));
		}
		else{
			O <- c(O,substr(se,z,z+2));
		}
		x <- x+1;
		z <- z+3;
	}
	
	N <- length(O);					#length of input sequence in codons
	
	A <- matrix(c(0.85,0.1399,0.0001,0.01,	# estimated transition probability matrix
		      0.05,0.93,0.0199,0.0001,
		      0.7,0.2998,0.0001,0.0001,
		      0.0001,0.9997,0.0001,0.0001),
		      4,4,byrow=TRUE);

	#nt <- c("A","C","T","G"); 			# nucleotide index
	#st <- c("cod","non-cod","start","stop");# state names
	
	#random chance of codon
	rand <- c(1:64);
	for(i in 1:64){ rand[i] <- 1/64;}
	
	#actual codons
	codons <- c("GGG","GGA","GGT","GGC","GAG","GAA","GAT","GAC","GTG","GTA","GTT","GTC","GCG","GCA","GCT","GCC",
		    "AGG","AGA","AGT","AGC","AAG","AAA","AAT","AAC","ATG","ATA","ATT","ATC","ACG","ACA","ACT","ACC",
   		    "TGG","TGA","TGT","TGC","TAG","TAA","TAT","TAC","TTG","TTA","TTT","TTC","TCG","TCA","TCT","TCC",
		    "CGG","CGA","CGT","CGC","CAG","CAA","CAT","CAC","CTG","CTA","CTT","CTC","CCG","CCA","CCT","CCC");

	#codon probabilities
	codprobs <- c(0.01876,0.01088,0.00914,0.0387,0.05016,0.01642,0.01227,0.03775,0.03868,0.00278,0.0041,0.01509,
		      0.01351,0.01012,0.01356,0.04264,0.01061,0.00618,0.00734,0.02427,0.04388,0.00976,0.00651,0.02322,
		      0.02232,0.00182,0.00653,0.02838,0.0084,0.00773,0.00742,0.03113,0.01351,0.00226,0.00674,0.01463,
		      0.00087,0.00095,0.00746,0.0216,0.00649,0.00158,0.00696,0.02854,0.00674,0.00342,0.00933,0.01986,
		      0.01267,0.00379,0.00435,0.02251,0.04187,0.00587,0.00485,0.01803,0.05978,0.00344,0.00493,0.02645,
		      0.00999,0.00945,0.01177,0.02923);
	
	#start codon probs
	startprobs <- c(0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,
			0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.9937,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,
			0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,
			0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001);

	#stop codon probs
	stopprobs <- c(0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,
		       0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,
		       0.0001,0.3979,0.0001,0.0001,0.2980,0.2980,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,
		       0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001,0.0001);

	# estimated emission probability matrix
	B <- matrix(c(codprobs,rand,startprobs,stopprobs),4,64,
	      	  byrow=TRUE, dimnames=(list(x=c("coding","non-coding","start","stop"),y=codons)));


	Pi <- c(0.09,0.9,0.005,0.005);		#estimated start state probabilities


   	lhoods<-c(1:iter); 				#record the likelihood at each iteration

   	for(x in 1:iter){
		res<-matrix(c(1:(S*S*(N-1)))); 	#gammas
		dim(res)<-c(S,S,(N-1));
	
		al<-forward(O,A,B,Pi,S,N); 		#forward alphas
		be<-backward(O,A,B,Pi,S,N); 		#backward betas	
	
		for(t in 1:(N-1)){			#calculate gammas
			for(i in 1:S){ 
				for(j in 1:S){
					res[i,j,t] <- (al[i,t]*A[i,j]*be[j,t+1]*B[j,O[t+1]]);
				}
			}
			tmp <- sum(res[,,t]);
			if(tmp <= 0)tmp <- 0.0001;
			res[,,t] <- (res[,,t] / tmp);
		}
	
		#record likelihood of current model
		lhoods[x] <- forwardbackward(al,be);	
	
		#adjust model parameters
		#new starting probabilities - Pi
		for(y in 1:S){
			Pi[y]<-0;
			for(z in 1:S){
				Pi[y]<-Pi[y]+res[y,z,1];
			}
		}

		#new transition matrix A
		for(p in 1:S){
			denom <- 0;
			for(l in 1:S){	
				for(t in 1:(N-1)){
					denom <- denom+res[p,l,t];
				}
			}
			if(denom <= 0)denom <- 0.0001;
			for(q in 1:S){
				num<-0;
				for(t in 1:(N-1)){
					num <- num+res[p,q,t];
				}
				#if(num <= 0) num <- 0.0001;
				A[p,q]= num / denom;
			}
		}

		#new observation probabilities matrix B
		for(j in 1:S){
			obsums <- matrix(c(1:length(codons)),1,length(codons),byrow=TRUE,dimnames=(list(x=1,y=codons)));
			gammasum <- sum(res[j,,]);
			if(gammasum <= 0)gammasum <- 0.0001;

			for(k in 1:length(codons)){
				obsums[1,k] <- 0;
			}

			for(t in 1:(N-1)){
				obsums[1,O[t]] <- obsums[1,O[t]] + sum(res[j,,t]);
			}		

			for(k in 1:length(codons)){
				B[j,k] <- (obsums[1,k] / gammasum);
			}
		}
	}

	#4)Viterbi to reconstruct best path
	pathprobs<-c(1:S);
	
	bestpaths<-matrix(c(1:(S*N)));
	dim(bestpaths)<-c(S,N);

	#initiation
	for(s in 1:S){
		pathprobs[s] <- Pi[s] * B[s,O[1]];
		bestpaths[s,1] <- (-1);
	}

	#induction
	for(t in 2:N){
		maxprobs<-c(1:S);
		for(i in 1:S){
			maxprobs[i] <- (-1);
			temp <- 0;
			for(j in 1:S){
				temp <- temp + (A[j,i] * pathprobs[j]);
				if(temp <= 0) temp <- 0.0001;
				if(temp > maxprobs[i]){
					maxprobs[i] <- temp;
					bestpaths[i,t] <- j;
				}
			}
		}
		for(j in 1:S){
			obsprob <- B[j,O[t]];
			if(obsprob <= 0) {obsprob <- 0.0001;}
			pathprobs[j] <- maxprobs[j] * obsprob;
		}
	}

	#termination
	maxprob <- (-1);
	bestpath <- c(1:N);
	for(i in 1:S){
		if(pathprobs[i] > maxprob){
			bestpath[N] <- i;
			maxprob <- pathprobs[i];
		}
	}

	#backtracking
	for(t in (N-1):1){
		bestpath[t] <- bestpaths[bestpath[t+1],(t+1)];
	}

	bestpath; 				#return reconstructed state sequence
	#ret<-c(lhoods,bestpath);
	#ret;
}

#paste in sequence and call emiter with desired number of iterations
testseq <- "ATGGGCCGCCGCCGCCGCCGCCGCCTAGTGATCCGCCGCCGCCTAG";
emiter(15,testseq);